@@ -24,16 +24,15 @@ from .. import mlog
 from .. import coredata
 from . import compilers
 from ..mesonlib import (
-    EnvironmentException, version_compare, Popen_safe, listify,
+    EnvironmentException, MesonException, version_compare, Popen_safe, listify,
     for_windows, for_darwin, for_cygwin, for_haiku, for_openbsd,
 )
+from .c_function_attributes import C_FUNC_ATTRIBUTES
 
 from .compilers import (
-    GCC_MINGW,
+    CompilerType,
     get_largefile_args,
     gnu_winlibs,
-    msvc_buildtype_args,
-    msvc_buildtype_linker_args,
     msvc_winlibs,
     vs32_instruction_set_args,
     vs64_instruction_set_args,
@@ -59,6 +58,13 @@ class CCompiler(Compiler):
     find_library_cache = {}
     internal_libs = gnu_compiler_internal_libs
 
+    @staticmethod
+    def attribute_check_func(name):
+        try:
+            return C_FUNC_ATTRIBUTES[name]
+        except KeyError:
+            raise MesonException('Unknown function attribute "{}"'.format(name))
+
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
         # If a child ObjC or CPP class has already set it, don't set it ourselves
         if not hasattr(self, 'language'):
@@ -67,10 +73,12 @@ class CCompiler(Compiler):
         self.id = 'unknown'
         self.is_cross = is_cross
         self.can_compile_suffixes.add('h')
-        if isinstance(exe_wrapper, str):
-            self.exe_wrapper = [exe_wrapper]
+        # If the exe wrapper was not found, pretend it wasn't set so that the
+        # sanity check is skipped and compiler checks use fallbacks.
+        if not exe_wrapper or not exe_wrapper.found():
+            self.exe_wrapper = None
         else:
-            self.exe_wrapper = exe_wrapper
+            self.exe_wrapper = exe_wrapper.get_command()
 
         # Set to None until we actually need to check this
         self.has_fatal_warnings_link_arg = None
@@ -111,9 +119,13 @@ class CCompiler(Compiler):
         return None, fname
 
     # The default behavior is this, override in MSVC
+    @functools.lru_cache(maxsize=None)
     def build_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
-        if self.id == 'clang' and self.clang_type == compilers.CLANG_OSX:
+        if getattr(self, 'compiler_type', False) and self.compiler_type.is_osx_compiler:
+            # Clang, GCC and ICC on macOS all use the same rpath arguments
             return self.build_osx_rpath_args(build_dir, rpath_paths, build_rpath)
+        elif self.compiler_type.is_windows_compiler:
+            return []
         return self.build_unix_rpath_args(build_dir, from_dir, rpath_paths, build_rpath, install_rpath)
 
     def get_dependency_gen_args(self, outtarget, outfile):
@@ -151,15 +163,8 @@ class CCompiler(Compiler):
         '''
         Get args for allowing undefined symbols when linking to a shared library
         '''
-        if self.id == 'clang':
-            if self.clang_type == compilers.CLANG_OSX:
-                # Apple ld
-                return ['-Wl,-undefined,dynamic_lookup']
-            else:
-                # GNU ld and LLVM lld
-                return ['-Wl,--allow-shlib-undefined']
-        elif self.id == 'gcc':
-            if self.gcc_type == compilers.GCC_OSX:
+        if self.id in ('clang', 'gcc'):
+            if self.compiler_type.is_osx_compiler:
                 # Apple ld
                 return ['-Wl,-undefined,dynamic_lookup']
             else:
@@ -214,7 +219,7 @@ class CCompiler(Compiler):
             p = Path(p)
             if p.exists():
                 paths.append(p.resolve().as_posix())
-        return paths
+        return tuple(paths)
 
     def get_compiler_dirs(self, env, name):
         '''
@@ -224,23 +229,19 @@ class CCompiler(Compiler):
         for line in stdo.split('\n'):
             if line.startswith(name + ':'):
                 return CCompiler._split_fetch_real_dirs(line.split('=', 1)[1])
-        return []
+        return ()
 
+    @functools.lru_cache()
     def get_library_dirs(self, env):
-        key = (tuple(self.exelist), env)
-        if key not in self.library_dirs_cache:
-            self.library_dirs_cache[key] = self.get_compiler_dirs(env, 'libraries')
-        return self.library_dirs_cache[key][:]
+        return self.get_compiler_dirs(env, 'libraries')
 
+    @functools.lru_cache()
     def get_program_dirs(self, env):
         '''
         Programs used by the compiler. Also where toolchain DLLs such as
         libstdc++-6.dll are found with MinGW.
         '''
-        key = (tuple(self.exelist), env)
-        if key not in self.program_dirs_cache:
-            self.program_dirs_cache[key] = self.get_compiler_dirs(env, 'programs')
-        return self.program_dirs_cache[key][:]
+        return self.get_compiler_dirs(env, 'programs')
 
     def get_pic_args(self):
         return ['-fPIC']
@@ -899,7 +900,7 @@ class CCompiler(Compiler):
             patterns += self._get_patterns(env, prefixes, stlibext, False)
         else:
             raise AssertionError('BUG: unknown libtype {!r}'.format(libtype))
-        return patterns
+        return tuple(patterns)
 
     @staticmethod
     def _sort_shlibs_openbsd(libs):
@@ -1045,10 +1046,23 @@ class CCompiler(Compiler):
             m = pattern.match(ret)
         return ret
 
+    def has_func_attribute(self, name, env):
+        # Just assume that if we're not on windows that dllimport and dllexport
+        # don't work
+        if not (for_windows(env.is_cross_build(), env) or
+                for_cygwin(env.is_cross_build(), env)):
+            if name in ['dllimport', 'dllexport']:
+                return False
+
+        # Clang and GCC both return warnings if the __attribute__ is undefined,
+        # so set -Werror
+        return self.compiles(self.attribute_check_func(name), env, extra_args='-Werror')
+
+
 class ClangCCompiler(ClangCompiler, CCompiler):
-    def __init__(self, exelist, version, clang_type, is_cross, exe_wrapper=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        ClangCompiler.__init__(self, clang_type)
+        ClangCompiler.__init__(self, compiler_type)
         default_warn_args = ['-Wall', '-Winvalid-pch']
         self.warn_args = {'1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
@@ -1074,7 +1088,7 @@ class ClangCCompiler(ClangCompiler, CCompiler):
 
     def get_linker_always_args(self):
         basic = super().get_linker_always_args()
-        if self.clang_type == compilers.CLANG_OSX:
+        if self.compiler_type.is_osx_compiler:
             return basic + ['-Wl,-headerpad_max_install_names']
         return basic
 
@@ -1108,9 +1122,9 @@ class ArmclangCCompiler(ArmclangCompiler, CCompiler):
 
 
 class GnuCCompiler(GnuCompiler, CCompiler):
-    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        GnuCompiler.__init__(self, gcc_type, defines)
+        GnuCompiler.__init__(self, compiler_type, defines)
         default_warn_args = ['-Wall', '-Winvalid-pch']
         self.warn_args = {'1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
@@ -1122,7 +1136,7 @@ class GnuCCompiler(GnuCompiler, CCompiler):
                                                        ['none', 'c89', 'c99', 'c11',
                                                         'gnu89', 'gnu99', 'gnu11'],
                                                        'none')})
-        if self.gcc_type == GCC_MINGW:
+        if self.compiler_type == CompilerType.GCC_MINGW:
             opts.update({
                 'c_winlibs': coredata.UserArrayOption('c_winlibs', 'Standard Win libraries to link against',
                                                       gnu_winlibs), })
@@ -1136,21 +1150,18 @@ class GnuCCompiler(GnuCompiler, CCompiler):
         return args
 
     def get_option_link_args(self, options):
-        if self.gcc_type == GCC_MINGW:
+        if self.compiler_type == CompilerType.GCC_MINGW:
             return options['c_winlibs'].value[:]
         return []
 
-    def get_std_shared_lib_link_args(self):
-        return ['-shared']
-
     def get_pch_use_args(self, pch_dir, header):
         return ['-fpch-preprocess', '-include', os.path.basename(header)]
 
 
 class ElbrusCCompiler(GnuCCompiler, ElbrusCompiler):
-    def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
-        GnuCCompiler.__init__(self, exelist, version, gcc_type, is_cross, exe_wrapper, defines, **kwargs)
-        ElbrusCompiler.__init__(self, gcc_type, defines)
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
+        GnuCCompiler.__init__(self, exelist, version, compiler_type, is_cross, exe_wrapper, defines, **kwargs)
+        ElbrusCompiler.__init__(self, compiler_type, defines)
 
     # It does support some various ISO standards and c/gnu 90, 9x, 1x in addition to those which GNU CC supports.
     def get_options(self):
@@ -1172,14 +1183,14 @@ class ElbrusCCompiler(GnuCCompiler, ElbrusCompiler):
 
 
 class IntelCCompiler(IntelCompiler, CCompiler):
-    def __init__(self, exelist, version, icc_type, is_cross, exe_wrapper=None, **kwargs):
+    def __init__(self, exelist, version, compiler_type, is_cross, exe_wrapper=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
-        IntelCompiler.__init__(self, icc_type)
+        IntelCompiler.__init__(self, compiler_type)
         self.lang_header = 'c-header'
         default_warn_args = ['-Wall', '-w3', '-diag-disable:remark', '-Wpch-messages']
         self.warn_args = {'1': default_warn_args,
                           '2': default_warn_args + ['-Wextra'],
-                          '3': default_warn_args + ['-Wextra', '-Wpedantic']}
+                          '3': default_warn_args + ['-Wextra']}
 
     def get_options(self):
         opts = CCompiler.get_options(self)
@@ -1199,12 +1210,6 @@ class IntelCCompiler(IntelCompiler, CCompiler):
             args.append('-std=' + std.value)
         return args
 
-    def get_std_shared_lib_link_args(self):
-        return ['-shared']
-
-    def has_arguments(self, args, env, code, mode):
-        return super().has_arguments(args + ['-diag-error', '10006'], env, code, mode)
-
 
 class VisualStudioCCompiler(CCompiler):
     std_warn_args = ['/W3']
@@ -1212,6 +1217,13 @@ class VisualStudioCCompiler(CCompiler):
     ignore_libs = gnu_compiler_internal_libs
     internal_libs = ()
 
+    crt_args = {'none': [],
+                'md': ['/MD'],
+                'mdd': ['/MDd'],
+                'mt': ['/MT'],
+                'mtd': ['/MTd'],
+                }
+
     def __init__(self, exelist, version, is_cross, exe_wrap, is_64):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
         self.id = 'msvc'
@@ -1221,7 +1233,7 @@ class VisualStudioCCompiler(CCompiler):
         self.warn_args = {'1': ['/W2'],
                           '2': ['/W3'],
                           '3': ['/W4']}
-        self.base_options = ['b_pch', 'b_ndebug'] # FIXME add lto, pgo and the like
+        self.base_options = ['b_pch', 'b_ndebug', 'b_vscrt'] # FIXME add lto, pgo and the like
         self.is_64 = is_64
 
     # Override CCompiler.get_always_args
@@ -1240,10 +1252,13 @@ class VisualStudioCCompiler(CCompiler):
         return ['/MDd']
 
     def get_buildtype_args(self, buildtype):
-        return msvc_buildtype_args[buildtype]
+        args = compilers.msvc_buildtype_args[buildtype]
+        if version_compare(self.version, '<18.0'):
+            args = [arg for arg in args if arg != '/Gw']
+        return args
 
     def get_buildtype_linker_args(self, buildtype):
-        return msvc_buildtype_linker_args[buildtype]
+        return compilers.msvc_buildtype_linker_args[buildtype]
 
     def get_pch_suffix(self):
         return 'pch'
@@ -1273,6 +1288,12 @@ class VisualStudioCCompiler(CCompiler):
             return ['/Fe' + target]
         return ['/Fo' + target]
 
+    def get_optimization_args(self, optimization_level):
+        return compilers.msvc_optimization_args[optimization_level]
+
+    def get_debug_args(self, is_debug):
+        return compilers.msvc_debug_args[is_debug]
+
     def get_dependency_gen_args(self, outtarget, outfile):
         return []
 
@@ -1291,12 +1312,20 @@ class VisualStudioCCompiler(CCompiler):
     def linker_to_compiler_args(self, args):
         return ['/link'] + args
 
-    def get_gui_app_args(self):
-        return ['/SUBSYSTEM:WINDOWS']
+    def get_gui_app_args(self, value):
+        # the default is for the linker to guess the subsystem based on presence
+        # of main or WinMain symbols, so always be explicit
+        if value:
+            return ['/SUBSYSTEM:WINDOWS']
+        else:
+            return ['/SUBSYSTEM:CONSOLE']
 
     def get_pic_args(self):
         return [] # PIC is handled by the loader on Windows
 
+    def gen_export_dynamic_link_args(self, env):
+        return [] # Not applicable with MSVC
+
     def get_std_shared_lib_link_args(self):
         return ['/DLL']
 
@@ -1447,6 +1476,30 @@ class VisualStudioCCompiler(CCompiler):
             return []
         return os.environ['INCLUDE'].split(os.pathsep)
 
+    def get_crt_compile_args(self, crt_val, buildtype):
+        if crt_val in self.crt_args:
+            return self.crt_args[crt_val]
+        assert(crt_val == 'from_buildtype')
+        # Match what build type flags used to do.
+        if buildtype == 'plain':
+            return []
+        elif buildtype == 'debug':
+            return self.crt_args['mdd']
+        elif buildtype == 'debugoptimized':
+            return self.crt_args['md']
+        elif buildtype == 'release':
+            return self.crt_args['md']
+        elif buildtype == 'minsize':
+            return self.crt_args['md']
+        else:
+            assert(buildtype == 'custom')
+            raise EnvironmentException('Requested C runtime based on buildtype, but buildtype is "custom".')
+
+    def has_func_attribute(self, name, env):
+        # MSVC doesn't have __attribute__ like Clang and GCC do, so just return
+        # false without compiling anything
+        return name in ['dllimport', 'dllexport']
+
 
 class ArmCCompiler(ArmCompiler, CCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
