@@ -15,21 +15,14 @@
 import configparser, os, platform, re, shlex, shutil, subprocess
 
 from . import coredata
-from .linkers import ArLinker, VisualStudioLinker
+from .linkers import ArLinker, ArmarLinker, VisualStudioLinker, DLinker
 from . import mesonlib
 from .mesonlib import EnvironmentException, Popen_safe
 from . import mlog
 
 from . import compilers
 from .compilers import (
-    CLANG_OSX,
-    CLANG_STANDARD,
-    CLANG_WIN,
-    GCC_CYGWIN,
-    GCC_MINGW,
-    GCC_OSX,
-    GCC_STANDARD,
-    ICC_STANDARD,
+    CompilerType,
     is_assembly,
     is_header,
     is_library,
@@ -211,16 +204,10 @@ def detect_cpu_family(compilers):
         trial = detect_windows_arch(compilers)
     else:
         trial = platform.machine().lower()
-
-    # Add mappings here as bugs are reported.
     if trial.startswith('i') and trial.endswith('86'):
         return 'x86'
     if trial.startswith('arm'):
         return 'arm'
-    if trial == 'mipsel':
-        return 'mips'
-    if trial == 'mips64el':
-        return 'mips64'
     if trial.startswith('ppc64'):
         return 'ppc64'
     if trial in ('amd64', 'x64'):
@@ -237,6 +224,7 @@ def detect_cpu_family(compilers):
                 # Ignore compilers that do not support has_builtin_define.
                 pass
         return 'x86_64'
+    # Add fixes here as bugs are reported.
 
     if trial not in known_cpu_families:
         mlog.warning('Unknown CPU family {!r}, please report this at '
@@ -318,10 +306,13 @@ class Environment:
             # Used by the regenchecker script, which runs meson
             self.coredata.meson_command = mesonlib.meson_command
             self.first_invocation = True
+        self.cross_info = None
+        self.exe_wrapper = None
         if self.coredata.cross_file:
             self.cross_info = CrossBuildInfo(self.coredata.cross_file)
-        else:
-            self.cross_info = None
+            if 'exe_wrapper' in self.cross_info.config['binaries']:
+                from .dependencies import ExternalProgram
+                self.exe_wrapper = ExternalProgram.from_cross_info(self.cross_info, 'exe_wrapper')
         self.cmd_line_options = options.cmd_line_options.copy()
 
         # List of potential compilers.
@@ -348,13 +339,11 @@ class Environment:
         # static libraries, and executables.
         # Versioning is added to these names in the backends as-needed.
         cross = self.is_cross_build()
-        if (not cross and mesonlib.is_windows()) \
-                or (cross and self.cross_info.has_host() and self.cross_info.config['host_machine']['system'] == 'windows'):
+        if mesonlib.for_windows(cross, self):
             self.exe_suffix = 'exe'
             self.object_suffix = 'obj'
             self.win_libdir_layout = True
-        elif (not cross and mesonlib.is_cygwin()) \
-                or (cross and self.cross_info.has_host() and self.cross_info.config['host_machine']['system'] == 'cygwin'):
+        elif mesonlib.for_cygwin(cross, self):
             self.exe_suffix = 'exe'
             self.object_suffix = 'o'
             self.win_libdir_layout = True
@@ -385,7 +374,7 @@ class Environment:
 
     def get_build_command(self, unbuffered=False):
         cmd = mesonlib.meson_command[:]
-        if unbuffered and 'python' in cmd[0]:
+        if unbuffered and 'python' in os.path.basename(cmd[0]):
             cmd.insert(1, '-u')
         return cmd
 
@@ -455,12 +444,12 @@ class Environment:
     def get_gnu_compiler_type(defines):
         # Detect GCC type (Apple, MinGW, Cygwin, Unix)
         if '__APPLE__' in defines:
-            return GCC_OSX
+            return CompilerType.GCC_OSX
         elif '__MINGW32__' in defines or '__MINGW64__' in defines:
-            return GCC_MINGW
+            return CompilerType.GCC_MINGW
         elif '__CYGWIN__' in defines:
-            return GCC_CYGWIN
-        return GCC_STANDARD
+            return CompilerType.GCC_CYGWIN
+        return CompilerType.GCC_STANDARD
 
     def warn_about_lang_pointing_to_cross(self, compiler_exe, evar):
         evar_str = os.environ.get(evar, 'WHO_WOULD_CALL_THEIR_COMPILER_WITH_THIS_NAME')
@@ -487,10 +476,7 @@ This is probably wrong, it should always point to the native compiler.''' % evar
             # Return value has to be a list of compiler 'choices'
             compilers = [compilers]
             is_cross = True
-            if self.cross_info.need_exe_wrapper():
-                exe_wrap = self.cross_info.config['binaries'].get('exe_wrapper', None)
-            else:
-                exe_wrap = []
+            exe_wrap = self.get_exe_wrapper()
         elif evar in os.environ:
             compilers = shlex.split(os.environ[evar])
             # Ensure ccache exists and remove it if it doesn't
@@ -567,14 +553,14 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                 if not defines:
                     popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
                     continue
-                gtype = self.get_gnu_compiler_type(defines)
+                compiler_type = self.get_gnu_compiler_type(defines)
                 if guess_gcc_or_lcc == 'lcc':
                     version = self.get_lcc_version_from_defines(defines)
                     cls = ElbrusCCompiler if lang == 'c' else ElbrusCPPCompiler
                 else:
                     version = self.get_gnu_version_from_defines(defines)
                     cls = GnuCCompiler if lang == 'c' else GnuCPPCompiler
-                return cls(ccache + compiler, version, gtype, is_cross, exe_wrap, defines, full_version=full_version)
+                return cls(ccache + compiler, version, compiler_type, is_cross, exe_wrap, defines, full_version=full_version)
 
             if 'armclang' in out:
                 # The compiler version is not present in the first line of output,
@@ -594,13 +580,13 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                 return cls(ccache + compiler, version, is_cross, exe_wrap, full_version=full_version)
             if 'clang' in out:
                 if 'Apple' in out or mesonlib.for_darwin(want_cross, self):
-                    cltype = CLANG_OSX
+                    compiler_type = CompilerType.CLANG_OSX
                 elif 'windows' in out or mesonlib.for_windows(want_cross, self):
-                    cltype = CLANG_WIN
+                    compiler_type = CompilerType.CLANG_MINGW
                 else:
-                    cltype = CLANG_STANDARD
+                    compiler_type = CompilerType.CLANG_STANDARD
                 cls = ClangCCompiler if lang == 'c' else ClangCPPCompiler
-                return cls(ccache + compiler, version, cltype, is_cross, exe_wrap, full_version=full_version)
+                return cls(ccache + compiler, version, compiler_type, is_cross, exe_wrap, full_version=full_version)
             if 'Microsoft' in out or 'Microsoft' in err:
                 # Latest versions of Visual Studio print version
                 # number to stderr but earlier ones print version
@@ -616,10 +602,15 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                 cls = VisualStudioCCompiler if lang == 'c' else VisualStudioCPPCompiler
                 return cls(compiler, version, is_cross, exe_wrap, is_64)
             if '(ICC)' in out:
-                # TODO: add microsoft add check OSX
-                inteltype = ICC_STANDARD
+                if mesonlib.for_darwin(want_cross, self):
+                    compiler_type = CompilerType.ICC_OSX
+                elif mesonlib.for_windows(want_cross, self):
+                    # TODO: fix ICC on Windows
+                    compiler_type = CompilerType.ICC_WIN
+                else:
+                    compiler_type = CompilerType.ICC_STANDARD
                 cls = IntelCCompiler if lang == 'c' else IntelCPPCompiler
-                return cls(ccache + compiler, version, inteltype, is_cross, exe_wrap, full_version=full_version)
+                return cls(ccache + compiler, version, compiler_type, is_cross, exe_wrap, full_version=full_version)
             if 'ARM' in out:
                 cls = ArmCCompiler if lang == 'c' else ArmCPPCompiler
                 return cls(ccache + compiler, version, is_cross, exe_wrap, full_version=full_version)
@@ -658,14 +649,14 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                     if not defines:
                         popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
                         continue
-                    gtype = self.get_gnu_compiler_type(defines)
+                    compiler_type = self.get_gnu_compiler_type(defines)
                     if guess_gcc_or_lcc == 'lcc':
                         version = self.get_lcc_version_from_defines(defines)
                         cls = ElbrusFortranCompiler
                     else:
                         version = self.get_gnu_version_from_defines(defines)
                         cls = GnuFortranCompiler
-                    return cls(compiler, version, gtype, is_cross, exe_wrap, defines, full_version=full_version)
+                    return cls(compiler, version, compiler_type, is_cross, exe_wrap, defines, full_version=full_version)
 
                 if 'G95' in out:
                     return G95FortranCompiler(compiler, version, is_cross, exe_wrap, full_version=full_version)
@@ -711,13 +702,13 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                 if not defines:
                     popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
                     continue
-                gtype = self.get_gnu_compiler_type(defines)
+                compiler_type = self.get_gnu_compiler_type(defines)
                 version = self.get_gnu_version_from_defines(defines)
-                return GnuObjCCompiler(ccache + compiler, version, gtype, is_cross, exe_wrap, defines)
+                return GnuObjCCompiler(ccache + compiler, version, compiler_type, is_cross, exe_wrap, defines)
             if out.startswith('Apple LLVM'):
-                return ClangObjCCompiler(ccache + compiler, version, CLANG_OSX, is_cross, exe_wrap)
+                return ClangObjCCompiler(ccache + compiler, version, CompilerType.CLANG_OSX, is_cross, exe_wrap)
             if out.startswith('clang'):
-                return ClangObjCCompiler(ccache + compiler, version, CLANG_STANDARD, is_cross, exe_wrap)
+                return ClangObjCCompiler(ccache + compiler, version, CompilerType.CLANG_STANDARD, is_cross, exe_wrap)
         self._handle_exceptions(popen_exceptions, compilers)
 
     def detect_objcpp_compiler(self, want_cross):
@@ -738,13 +729,13 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                 if not defines:
                     popen_exceptions[' '.join(compiler)] = 'no pre-processor defines'
                     continue
-                gtype = self.get_gnu_compiler_type(defines)
+                compiler_type = self.get_gnu_compiler_type(defines)
                 version = self.get_gnu_version_from_defines(defines)
-                return GnuObjCPPCompiler(ccache + compiler, version, gtype, is_cross, exe_wrap, defines)
+                return GnuObjCPPCompiler(ccache + compiler, version, compiler_type, is_cross, exe_wrap, defines)
             if out.startswith('Apple LLVM'):
-                return ClangObjCPPCompiler(ccache + compiler, version, CLANG_OSX, is_cross, exe_wrap)
+                return ClangObjCPPCompiler(ccache + compiler, version, CompilerType.CLANG_OSX, is_cross, exe_wrap)
             if out.startswith('clang'):
-                return ClangObjCPPCompiler(ccache + compiler, version, CLANG_STANDARD, is_cross, exe_wrap)
+                return ClangObjCPPCompiler(ccache + compiler, version, CompilerType.CLANG_STANDARD, is_cross, exe_wrap)
         self._handle_exceptions(popen_exceptions, compilers)
 
     def detect_java_compiler(self):
@@ -820,6 +811,8 @@ This is probably wrong, it should always point to the native compiler.''' % evar
         # up to date language version at time (2016).
         if 'DC' in os.environ:
             exelist = shlex.split(os.environ['DC'])
+            if os.path.basename(exelist[-1]).startswith(('ldmd', 'gdmd')):
+                    raise EnvironmentException('Meson doesn\'t support %s as it\'s only a DMD frontend for another compiler. Please provide a valid value for DC or unset it so that Meson can resolve the compiler by itself.' % exelist[-1])
         elif self.is_cross_build() and want_cross:
             exelist = mesonlib.stringlistify(self.cross_info.config['binaries']['d'])
             is_cross = True
@@ -840,12 +833,23 @@ This is probably wrong, it should always point to the native compiler.''' % evar
             raise EnvironmentException('Could not execute D compiler "%s"' % ' '.join(exelist))
         version = search_version(out)
         full_version = out.split('\n', 1)[0]
+
+        # Detect the target architecture, required for proper architecture handling on Windows.
+        c_compiler = {}
+        is_msvc = mesonlib.is_windows() and 'VCINSTALLDIR' in os.environ
+        if is_msvc:
+            c_compiler = {'c': self.detect_c_compiler(want_cross)} # MSVC compiler is required for correct platform detection.
+
+        arch = detect_cpu_family(c_compiler)
+        if is_msvc and arch == 'x86':
+            arch = 'x86_mscoff'
+
         if 'LLVM D compiler' in out:
-            return compilers.LLVMDCompiler(exelist, version, is_cross, full_version=full_version)
+            return compilers.LLVMDCompiler(exelist, version, is_cross, arch, full_version=full_version)
         elif 'gdc' in out:
-            return compilers.GnuDCompiler(exelist, version, is_cross, full_version=full_version)
+            return compilers.GnuDCompiler(exelist, version, is_cross, arch, full_version=full_version)
         elif 'The D Language Foundation' in out or 'Digital Mars' in out:
-            return compilers.DmdDCompiler(exelist, version, is_cross, full_version=full_version)
+            return compilers.DmdDCompiler(exelist, version, is_cross, arch, full_version=full_version)
         raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
 
     def detect_swift_compiler(self):
@@ -877,6 +881,12 @@ This is probably wrong, it should always point to the native compiler.''' % evar
             elif isinstance(compiler, compilers.ClangCompiler):
                 # Use llvm-ar if available; needed for LTO
                 linkers = [self.clang_static_linker, self.default_static_linker]
+            elif isinstance(compiler, compilers.DCompiler):
+                # Prefer static linkers over linkers used by D compilers
+                if mesonlib.is_windows():
+                    linkers = [self.vs_static_linker, compiler.get_linker_exelist()]
+                else:
+                    linkers = [self.default_static_linker, compiler.get_linker_exelist()]
             else:
                 linkers = [self.default_static_linker]
         popen_exceptions = {}
@@ -892,6 +902,14 @@ This is probably wrong, it should always point to the native compiler.''' % evar
                 continue
             if '/OUT:' in out or '/OUT:' in err:
                 return VisualStudioLinker(linker)
+            if p.returncode == 0 and ('armar' in linker or 'armar.exe' in linker):
+                return ArmarLinker(linker)
+            if 'DMD32 D Compiler' in out or 'DMD64 D Compiler' in out:
+                return DLinker(linker, compiler.arch)
+            if 'LDC - the LLVM D compiler' in out:
+                return DLinker(linker, compiler.arch)
+            if 'GDC' in out and ' based on D ' in out:
+                return DLinker(linker, compiler.arch)
             if p.returncode == 0:
                 return ArLinker(linker)
             if p.returncode == 1 and err.startswith('usage'): # OSX
@@ -980,16 +998,21 @@ This is probably wrong, it should always point to the native compiler.''' % evar
         out = out.split('\n')[index].lstrip('libraries: =').split(':')
         return [os.path.normpath(p) for p in out]
 
+    def get_exe_wrapper(self):
+        if not self.cross_info.need_exe_wrapper():
+            from .dependencies import EmptyExternalProgram
+            return EmptyExternalProgram()
+        return self.exe_wrapper
+
+
 class CrossBuildInfo:
     def __init__(self, filename):
         self.config = {'properties': {}}
         self.parse_datafile(filename)
-        if 'target_machine' in self.config:
-            return
-        if 'host_machine' not in self.config:
+        if 'host_machine' not in self.config and 'target_machine' not in self.config:
             raise mesonlib.MesonException('Cross info file must have either host or a target machine.')
-        if 'binaries' not in self.config:
-            raise mesonlib.MesonException('Cross file is missing "binaries".')
+        if 'host_machine' in self.config and 'binaries' not in self.config:
+            raise mesonlib.MesonException('Cross file with "host_machine" is missing "binaries".')
 
     def ok_type(self, i):
         return isinstance(i, (str, int, bool))
@@ -1038,6 +1061,12 @@ class CrossBuildInfo:
     def get_stdlib(self, language):
         return self.config['properties'][language + '_stdlib']
 
+    def get_host_system(self):
+        "Name of host system like 'linux', or None"
+        if self.has_host():
+            return self.config['host_machine']['system']
+        return None
+
     def get_properties(self):
         return self.config['properties']
 
